import copy
import json

from jsonizer import Jsonizer
from tonclient.types import AbiContract, AbiFunction, AbiEvent, AbiData, AbiParam


def file_to_abi(filename: str) -> AbiContract:
    with open(filename, 'r') as file:
        data = json.load(file)
    return json_to_abi(data, mutate=True)


def json_to_abi(data: dict, mutate: bool = False) -> AbiContract:
    # fix abi events - removing "outputs" field
    if len(data['events']) > 0:
        if not mutate:
            data = copy.deepcopy(data)
        for event in data['events']:
            event.pop('outputs', None)
    parser = Jsonizer(AbiContract, AbiFunction, AbiEvent, AbiData, AbiParam, lowercase_keys=True, replace_space='_')
    return parser.parse(data)


def abi_params_to_str(abi_params: list[AbiParam]) -> str:
    result = list()
    for abi_param in abi_params:
        sub_result = _abi_param_to_tuple(abi_param)
        result.append(sub_result)
    return ','.join(result)


def _abi_param_to_tuple(abi_param: AbiParam) -> str:
    type_ = abi_param.type
    if abi_param.components:
        components_str = abi_params_to_str(abi_param.components)
        type_ = type_.replace('tuple', '(' + components_str + ')')
    return type_



from typing import Type

import attrs


def convert_nested(data: dict, data_class: Type[attrs.define]):
    for field in attrs.fields(data_class):
        value = data.get(field.name)
        if isinstance(value, dict) and field.type is not dict:
            data[field.name] = field.type(**value)



import re

from tvmbase.constants import ADDRESS_REGEX, IDX_REGEX


def is_address(value: str) -> bool:
    return bool(re.fullmatch(ADDRESS_REGEX, value))


def is_idx(value: str) -> bool:
    return bool(re.fullmatch(IDX_REGEX, value))


def decode_bytes(value: str) -> str:
    return bytes.fromhex(value).decode()



class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        key = hash((cls, str(args), str(kwargs)))
        if key not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[key] = instance
        return cls._instances[key]



from tonclient.types import ParamsOfParse, ParamsOfQueryCollection

from tvmbase.client import Client
from tvmbase.models.data import TransactionData
from tvmbase.models.tvm.base import BaseTvm
from tvmbase.utils.attrs_helper import convert_nested


class Transaction(BaseTvm):
    DATA_TYPE = TransactionData

    def __init__(self, idx: str, data: TransactionData):
        super().__init__(idx)
        self.data = data

    @staticmethod
    def gql_query(idx: str) -> ParamsOfQueryCollection:
        return ParamsOfQueryCollection(
            collection='transactions',
            result='boc',
            limit=1,
            filter={'id': {'eq': idx}},
        )

    @classmethod
    async def from_boc(cls, client: Client, boc: str, **kwargs) -> 'Transaction':
        parse_params = ParamsOfParse(boc=boc)
        parsed = await client.boc.parse_transaction(params=parse_params)
        parsed_dict = parsed.parsed
        convert_nested(parsed_dict, TransactionData)
        idx = parsed.parsed['id']
        data = TransactionData(**parsed_dict, **kwargs)
        return cls(idx, data)



from enum import Enum

from tvmbase.models.tvm.account import Account
from tvmbase.models.tvm.message import Message
from tvmbase.models.tvm.transaction import Transaction


class BocType(Enum):
    ACCOUNT = Account
    MESSAGE = Message
    TRANSACTION = Transaction

    def __str__(self) -> str:
        return self.name

    __repr__ = __str__



from tonclient.types import ParamsOfParse, ParamsOfQueryCollection, ResultOfQueryCollection

from tvmbase.client import Client
from tvmbase.models.data import MessageData
from tvmbase.models.tvm.base import BaseTvm


class Message(BaseTvm):
    DATA_TYPE: type = MessageData

    def __init__(self, idx: str, data: MessageData):
        super().__init__(idx)
        self.data = data

    @staticmethod
    def gql_query(idx: str) -> ParamsOfQueryCollection:
        return ParamsOfQueryCollection(
            collection='messages',
            result='boc created_at dst_transaction { id } src_transaction { id }',
            limit=1,
            filter={'id': {'eq': idx}},
        )

    @classmethod
    async def from_query_result(cls, client: Client, idx: str, result: ResultOfQueryCollection) -> 'Message':
        result_dict = result.result[0]
        boc = result_dict['boc']
        created_at = result_dict['created_at']
        dst_transaction_id = cls._get_id(result_dict, 'dst_transaction')
        src_transaction_id = cls._get_id(result_dict, 'src_transaction')
        return await cls.from_boc(
            client,
            boc,
            created_at=created_at,
            dst_transaction_id=dst_transaction_id,
            src_transaction_id=src_transaction_id,
        )

    @staticmethod
    def _get_id(result_dict: dict, key: str) -> str:
        data = result_dict.get(key)
        if isinstance(data, dict):
            return data.get('id')

    @classmethod
    async def from_boc(cls, client: Client, boc: str, **kwargs) -> 'Message':
        parse_params = ParamsOfParse(boc=boc)
        parsed = await client.boc.parse_message(params=parse_params)
        idx = parsed.parsed['id']
        created_at = parsed.parsed.pop('created_at', None)
        assert 'created_at' not in kwargs or created_at is None or created_at == kwargs['created_at'], \
            f'Different created at time in message {idx}'
        data = MessageData(**parsed.parsed, **kwargs)
        return cls(idx, data)



from dataclasses import dataclass

from tvmbase.exceptions import NetworkException
from tvmbase.utils.singleton import SingletonMeta


@dataclass(frozen=True, slots=True)
class Network:
    name: str
    endpoints: list[str]
    everlive_domain: str
    everscan_domain: str


class NetworkFactory(metaclass=SingletonMeta):
    custom_networks: dict[str, Network] = dict()

    def __init__(self, evercloud_key: str):
        self.evercloud_key = evercloud_key

    def mainnet(self) -> Network:
        return Network(
            name='main',
            endpoints=[f'https://mainnet.evercloud.dev/{self.evercloud_key}/graphql'],
            everlive_domain='ever.live',
            everscan_domain='everscan.io',
        )

    def devnet(self) -> Network:
        return Network(
            name='dev',
            endpoints=[f'https://devnet.evercloud.dev/{self.evercloud_key}/graphql'],
            everlive_domain='ever.live',
            everscan_domain='everscan.io',
        )

    def from_name(self, name: str) -> Network:
        name = name.lower()
        match name:
            case 'main' | 'mainnet':
                return self.mainnet()
            case 'dev' | 'devnet':
                return self.devnet()
        if name in self.custom_networks:
            return self.custom_networks.get(name)
        raise NetworkException(f'Unknown network "{name}"')

    @staticmethod
    def add_custom(network: Network):
        NetworkFactory.custom_networks[network.name] = network

    def get_all(self) -> list[Network]:
        return [self.mainnet(), self.devnet()] + list(self.custom_networks.values())
